﻿/*
   Copyright 2019 Nils Kopal, nils.kopal<at>cryptool.org
                  Christian Bender, christian.bender<at>cryptool.org

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
*/

using System;

namespace MysteryCipher.Ciphers
{
    public class MysteryCipher1
    {
        #region sbox definitions
        // SBox (4bit):
        public static readonly int[] SBox4 = new int[] { 10, 05, 15, 08, 11, 00, 03, 07, 01, 13, 09, 12, 06, 14, 02, 04 };
        // SBox reversed (4bit):
        public static readonly int[] SBox4r = new int[]{ 05, 08, 14, 06, 15, 01, 12, 07, 03, 10, 00, 04, 11, 09, 13, 02 };
        #endregion

        /// <summary>
        /// Encrypts a plaintext using the given roundkeys
        /// </summary>
        /// <param name="plaintext">16bit</param>
        /// <param name="roundkeys">4x 16bit</param>
        /// <returns></returns>
        public int Encrypt(int plaintext, int[] roundkeys)
        {
            if(plaintext > 0xffff)
            {
                throw new Exception("Given plaintext is > 0xff. Blocksize is 16bit only");
            }
            int ciphertext = plaintext & 0xffff;
            for (var round = 0; round < roundkeys.Length - 1; round++)
            {
                ciphertext = EncryptSingleRound(ciphertext, roundkeys[round], round == roundkeys.Length - 2);
            }
            ciphertext = (ciphertext ^ (roundkeys[roundkeys.Length - 1] & 0xffff));
            return ciphertext;
        }

        /// <summary>
        /// Decrypts a plaintext using the given roundkeys
        /// <param name="ciphertext">16bit</param>
        /// <param name="roundkeys">4x 16bit</param>
        /// <returns></returns>
        public int Decrypt(int ciphertext, int[] roundkeys)
        {
            if (ciphertext > 0xffff)
            {
                throw new Exception("Given ciphertext is > 0xff. Blocksize is 16bit only");
            }
            int plaintext = ciphertext & 0xffff;
            plaintext = (plaintext ^ (roundkeys[roundkeys.Length - 1] & 0xffff));
            for (var round = roundkeys.Length - 2; round >= 0; round--)
            {
                plaintext = DecryptSingleRound(plaintext, roundkeys[round], round == roundkeys.Length - 2);
            }
            return plaintext;
        }

        /// <summary>
        /// Encrypts a single round using a single round key
        /// </summary>
        /// <param name="plaintext">16bit</param>
        /// <param name="roundkey">16bit</param>
        /// <param name="lastRound"></param>
        /// <returns></returns>
        private int EncryptSingleRound(int plaintext, int roundkey, bool lastRound = false)
        {
            //step 1: XOR roundkey
            var ciphertext = (plaintext ^ (roundkey & 0xffff));

            //step 1: do substitution with 4 sboxes in parallel
            ciphertext = (SBox4[0xf & (ciphertext >> 12)] << 12 |
                          SBox4[0xf & (ciphertext >> 8)] << 8 |
                          SBox4[0xf & (ciphertext >> 4)] << 4 |
                          SBox4[0xf & ciphertext]);

            //step 2: do permutation
            if (!lastRound)
            {
                var bit15 = 1 & (ciphertext >> 15);
                var bit14 = 1 & (ciphertext >> 14);
                var bit13 = 1 & (ciphertext >> 13);
                var bit12 = 1 & (ciphertext >> 12);
                var bit11 = 1 & (ciphertext >> 11);
                var bit10 = 1 & (ciphertext >> 10);
                var bit9 = 1 & (ciphertext >> 9);
                var bit8 = 1 & (ciphertext >> 8);
                var bit7 = 1 & (ciphertext >> 7);
                var bit6 = 1 & (ciphertext >> 6);
                var bit5 = 1 & (ciphertext >> 5);
                var bit4 = 1 & (ciphertext >> 4);
                var bit3 = 1 & (ciphertext >> 3);
                var bit2 = 1 & (ciphertext >> 2);
                var bit1 = 1 & (ciphertext >> 1);
                var bit0 = 1 & (ciphertext >> 0);

                ciphertext = ((bit0 << 6) |
                              (bit1 << 13) |
                              (bit2 << 2) |
                              (bit3 << 11) |
                              (bit4 << 9) |
                              (bit5 << 10) |
                              (bit6 << 7) |
                              (bit7 << 4) |
                              (bit8 << 8) |
                              (bit9 << 14) |
                              (bit10 << 12) |
                              (bit11 << 3) |
                              (bit12 << 15) |
                              (bit13 << 1) |
                              (bit14 << 0) |
                              (bit15 << 5));
            }
            return ciphertext;
        }

        /// <summary>
        /// Decrypts a single round using a single round key
        /// </summary>
        /// <param name="ciphertext">16bit</param>
        /// <param name="roundkey">16bit</param>
        /// <param name="lastRound"></param>
        /// <returns></returns>
        private int DecryptSingleRound(int ciphertext, int roundkey, bool lastRound = false)
        {
            int plaintext = ciphertext;

            //step 1: reverse permutation
            if (!lastRound)
            {
                var bit15 = 1 & (ciphertext >> 15);
                var bit14 = 1 & (ciphertext >> 14);
                var bit13 = 1 & (ciphertext >> 13);
                var bit12 = 1 & (ciphertext >> 12);
                var bit11 = 1 & (ciphertext >> 11);
                var bit10 = 1 & (ciphertext >> 10);
                var bit9 = 1 & (ciphertext >> 9);
                var bit8 = 1 & (ciphertext >> 8);
                var bit7 = 1 & (ciphertext >> 7);
                var bit6 = 1 & (ciphertext >> 6);
                var bit5 = 1 & (ciphertext >> 5);
                var bit4 = 1 & (ciphertext >> 4);
                var bit3 = 1 & (ciphertext >> 3);
                var bit2 = 1 & (ciphertext >> 2);
                var bit1 = 1 & (ciphertext >> 1);
                var bit0 = 1 & (ciphertext >> 0);

                plaintext = ((bit0 << 14) |
                             (bit1 << 13) |
                             (bit2 << 2) |
                             (bit3 << 11) |
                             (bit4 << 7) |
                             (bit5 << 15) |
                             (bit6 << 0) |
                             (bit7 << 6) |
                             (bit8 << 8) |
                             (bit9 << 4) |
                             (bit10 << 5) |
                             (bit11 << 3) |
                             (bit12 << 10) |
                             (bit13 << 1) |
                             (bit14 << 9) |
                             (bit15 << 12));
            }

            //step 2: reverse substitution with reversed sbox
            plaintext = (SBox4r[0xf & (plaintext >> 12)] << 12 |
                         SBox4r[0xf & (plaintext >> 8)] << 8 |
                         SBox4r[0xf & (plaintext >> 4)] << 4 |
                         SBox4r[0xf & plaintext]);

            //step 3: XOR roundkey
            plaintext = (plaintext ^ (roundkey & 0xffff));
            return plaintext;
        }

        /// <summary>
        /// Encrypts a given byte array using MysteryCipher 1 in CBC mode
        /// IV is per default = 0x00
        /// </summary>
        /// <param name="plaintext"></param>
        /// <param name="roundkeys"></param>
        /// <returns></returns>
        public byte[] Encrypt(byte[] plaintext, int[] roundkeys)
        {
            //append plaintext with 0 if too short
            if(plaintext.Length % 2 == 1)
            {
                byte[] newPlaintext = new byte[plaintext.Length + 1];
                Array.Copy(plaintext, newPlaintext, plaintext.Length);
                plaintext = newPlaintext;
            }            
            byte[] ciphertext = new byte[plaintext.Length];
            UInt16 cold = 0;
            for(int offset = 0; offset < plaintext.Length; offset+=2)
            {
                UInt16 p = BitConverter.ToUInt16(plaintext, offset);
                p = (UInt16)(p ^ cold);
                UInt16 c = (UInt16)Encrypt(p, roundkeys);
                cold = c;
                byte[] c_bytes = BitConverter.GetBytes(c);
                ciphertext[offset] = c_bytes[0];
                ciphertext[offset + 1] = c_bytes[1];
            }
            return ciphertext;
        }

        /// <summary>
        /// Decrypts a given plaintext  using MysteryCipher 1 in CBC mode
        /// IV is per default = 0x00
        /// </summary>
        /// <param name="ciphertext"></param>
        /// <param name="roundkeys"></param>
        /// <returns></returns>
        public byte[] Decrypt(byte[] ciphertext, int[] roundkeys)
        {
            if(ciphertext.Length % 2 == 1)
            {
                throw new ArgumentException("Ciphertext must be of even length");
            }
            byte[] plaintext = new byte[ciphertext.Length];
            UInt16 cold = 0;
            for (int offset = 0; offset < plaintext.Length; offset += 2)
            {
                UInt16 c = BitConverter.ToUInt16(ciphertext, offset);                
                UInt16 p = (UInt16)Decrypt(c, roundkeys);
                p = (UInt16)(p ^ cold);
                cold = c;
                byte[] p_bytes = BitConverter.GetBytes(p);
                plaintext[offset] = p_bytes[0];
                plaintext[offset + 1] = p_bytes[1];
            }
            return plaintext;
        }
    }
}
